import os
import math
import yaml
import typing
import pickle
import miditoolkit
from copy import deepcopy
from remi_z import keys_normalization


class MidiEncoder(object):
    def __init__(self):
        self.DEFAULT_TS = (4, 4)
        self.DEFAULT_TEMPO = 120.0

    def time_to_pos(self, t, ticks_per_beat):
        pos_resolution = 12 # 12 positions per beat
        ret = round(t * pos_resolution / ticks_per_beat)
        return ret

    def reduce_time_signature(self, numerator, denominator):
        max_ts_denominator = 2 ** 6
        while denominator > max_ts_denominator and denominator % 2 == 0 and numerator % 2 == 0:
            denominator //= 2
            numerator //= 2
        # decomposition (when length of a bar exceed max_notes_per_bar)
        max_notes_per_bar = 2
        while numerator > max_notes_per_bar * denominator:
            for i in range(2, numerator + 1):
                if numerator % i == 0:
                    numerator //= i
                    break
        return numerator, denominator

    def collect_pos_info(self, midi_obj, trunc_pos=None, tracks=None, remove_same_notes=False, end_offset=0):
        if tracks is not None:
            from collections.abc import Iterable
            assert isinstance(tracks, (int, Iterable))
            if isinstance(tracks, str):
                tracks = int(tracks)
            if isinstance(tracks, int):
                if tracks < 0:
                    tracks = len(midi_obj.instruments) + tracks
                tracks = (tracks,)

        max_pos = 0
        for inst in midi_obj.instruments:
            for note in inst.notes:
                pos = self.time_to_pos(note.start, midi_obj.ticks_per_beat)
                max_pos = max(max_pos, pos)
        max_pos = max_pos + 1  # 最大global position
        if trunc_pos is not None:
            max_pos = min(max_pos, trunc_pos)

        pos_info = [
            [None, None, None, None, None]  # (bar, ts, local_pos, tempo, insts_notes)
            for _ in range(max_pos)
        ]
        pos_info: typing.List
        # bar: every pos
        # ts: only at pos where it changes, otherwise None
        # local_pos: every pos
        # tempo: only at pos where it changes, otherwise None
        # insts_notes: only at pos where the note starts, otherwise None

        ts_changes = midi_obj.time_signature_changes
        zero_pos_ts_change = False
        for ts_change in ts_changes:
            pos = self.time_to_pos(ts_change.time, midi_obj.ticks_per_beat)
            if pos >= max_pos:
                continue
            if pos == 0:
                zero_pos_ts_change = True
            ts_numerator = int(ts_change.numerator)
            ts_denominator = int(ts_change.denominator)
            
            ts_numerator, ts_denominator = self.reduce_time_signature(ts_numerator, ts_denominator)
            pos_info[pos][1] = (ts_numerator, ts_denominator)
        if not zero_pos_ts_change:
            pos_info[0][1] = self.DEFAULT_TS

        tempo_changes = midi_obj.tempo_changes
        zero_pos_tempo_change = False
        for tempo_change in tempo_changes:
            pos = self.time_to_pos(tempo_change.time, midi_obj.ticks_per_beat)
            if pos >= max_pos:
                continue
            if pos == 0:
                zero_pos_tempo_change = True
            pos_info[pos][3] = tempo_change.tempo
        if not zero_pos_tempo_change:
            pos_info[0][3] = self.DEFAULT_TEMPO

        insts = midi_obj.instruments
        for inst_idx, inst in enumerate(insts):
            if tracks is not None and inst_idx not in tracks:
                continue
            # if self.ignore_insts:
            #     inst_id = 0
            # else:
            inst_id = 128 if inst.is_drum else int(inst.program)
            notes = inst.notes
            for note in notes:
                pitch = int(note.pitch)
                velocity = int(note.velocity)
                start_time = int(note.start)
                end_time = int(note.end + end_offset)
                # assert end_time > start_time
                if end_time < start_time:
                    print('Warning: end_time < start_time')
                pos_start = self.time_to_pos(start_time, midi_obj.ticks_per_beat)
                pos_end = self.time_to_pos(end_time, midi_obj.ticks_per_beat)
                duration = pos_end - pos_start

                if pos_info[pos_start][4] is None:
                    pos_info[pos_start][4] = dict()
                if inst_id not in pos_info[pos_start][4]:
                    pos_info[pos_start][4][inst_id] = []
                note_info = [pitch, duration, velocity]
                if remove_same_notes:
                    if note_info in pos_info[pos_start][4][inst_id]:
                        continue
                pos_info[pos_start][4][inst_id].append([pitch, duration, velocity])

        cnt = 0
        bar = 0
        measure_length = None
        ts = self.DEFAULT_TS  # default MIDI time signature
        for j in range(max_pos):
            now_ts = pos_info[j][1]
            if now_ts is not None:
                if now_ts != ts:
                    ts = now_ts
            if cnt == 0:
                beat_note_factor = 4
                pos_resolution = 12
                measure_length = ts[0] * beat_note_factor * pos_resolution // ts[1]
            pos_info[j][0] = bar
            pos_info[j][2] = cnt
            cnt += 1
            if cnt >= measure_length:
                assert cnt == measure_length, 'invalid time signature change: pos = {}'.format(j)
                cnt = 0
                bar += 1

        return pos_info

    def normalize_pitch(self, pos_info):
        assert self.key_profile is not None, "Please load key_profile first, using load_key_profile method."

        pitch_shift, is_major, _, _ = keys_normalization.get_pitch_shift(
            pos_info,
            self.key_profile,
            normalize=True, use_duration=True, use_velocity=True,
            ensure_valid_range=True
        )
        pitch_shift = int(pitch_shift)
      
        return is_major, pitch_shift





def load_midi(file_path=None, file=None, midi_checker='default'):
    """
    Open and check MIDI file, return MIDI object by miditoolkit.
    :param file_path:
    :param midi_checker:
    :return:
    """
    midi_obj = miditoolkit.midi.parser.MidiFile(filename=file_path, file=file)

    if midi_checker is not None and midi_checker != 'none':
        if isinstance(midi_checker, str):
            if midi_checker == 'default':
                midi_checker = default_check_midi
            else:
                raise ValueError("midi checker does not support value: %s" % midi_checker)

        midi_checker(midi_obj)

    return midi_obj

def default_check_midi(midi_obj):
    # check abnormal values in parse result
    max_time_length = 2 ** 31
    assert all(0 <= j.start < max_time_length
               and 0 <= j.end < max_time_length
               for i in midi_obj.instruments for j in i.notes), 'Bad note time'
    assert all(0 < j.numerator < max_time_length and 0 < j.denominator < max_time_length for j in
               midi_obj.time_signature_changes), 'Bad time signature value'
    assert 0 < midi_obj.ticks_per_beat < max_time_length, 'Bad ticks per beat'

    midi_notes_count = sum(len(inst.notes) for inst in midi_obj.instruments)
    assert midi_notes_count > 0, 'Blank note.'


def fill_pos_ts_and_tempo_(pos_info):
    '''
    Original item of pos_info: [bar, ts, local_pos, tempo, insts_notes]
    '''
    cur_ts = pos_info[0][1]
    cur_tempo = pos_info[0][3]
    assert cur_ts is not None
    assert cur_tempo is not None
    for idx in range(len(pos_info)):
        pos_item = pos_info[idx]
        if pos_item[1] is not None:
            cur_ts = pos_item[1]
        if pos_item[3] is not None:
            cur_tempo = pos_item[3]
        if pos_item[2] == 0:
            if pos_item[1] is None:
                pos_item[1] = cur_ts
            if pos_item[3] is None:
                pos_item[3] = cur_tempo
    return pos_info


def convert_tempo_to_id(x):
    '''
    Return: e, int, tempo id.
    '''
    x = max(x, 16) # min_tempo = 16
    x = min(x, 256) # max_tempo = 256
    tempo_quant = 12
    x = x / 16
    e = round(math.log2(x) * tempo_quant)
    return e


def convert_id_to_tempo(x):
    tempo_quant = 12
    min_tempo = 16
    return 2 ** (x / tempo_quant) * min_tempo






import os
import sys
dirof = os.path.dirname
sys.path.insert(0, dirof(__file__))

from utils import read_yaml


def main():
    # tide_up_time_signature()
    # test_tempo_converter()
    # generate_tempo_dict()
    test_ts_tempo_converter()


def test_ts_tempo_converter():
    ts = [(4, 4), (3, 4), (2, 4), (6, 8), (9, 8), (12, 8)]
    for t in ts:
        print(t, convert_time_signature_to_ts_token(t[0], t[1]))

    tempo = [60, 88, 96, 108, 128, 180, 200]
    for t in tempo:
        print(t, convert_tempo_to_tempo_token(t))


class TimeSignatureUtil:
    @classmethod
    def convert_time_signature_to_ts_token(cls, numerator, denominator):
        if hasattr(cls, 'ts_dict'):
            data = cls.ts_dict  
        else:
            ts_fp = os.path.join(os.path.dirname(__file__), 'dict_time_signature.yaml')
            data = read_yaml(ts_fp)
            cls.ts_dict = data

        valid = False
        for k, v in data.items():
            if v == '({}, {})'.format(numerator, denominator):
                valid = True
                return k
            
        if not valid:
            raise ValueError('Invalid time signature: {}/{}'.format(numerator, denominator))
        
    @classmethod
    def convert_time_signature_token_to_tuple(cls, token):
        if hasattr(cls, 'ts_dict'):
            data = cls.ts_dict  
        else:
            ts_fp = os.path.join(os.path.dirname(__file__), 'dict_time_signature.yaml')
            data = read_yaml(ts_fp)
            cls.ts_dict = data

        if token in data:
            ret = data[token]
            t1, t2 = ret[1:-1].split(',')
            ret = int(t1), int(t2)
            return ret
        else:
            raise ValueError('Invalid time signature token: {}'.format(token))


def convert_time_signature_to_ts_token(numerator, denominator):
    ts_fp = os.path.join(os.path.dirname(__file__), 'dict_time_signature.yaml')
    data = read_yaml(ts_fp)
    valid = False
    for k, v in data.items():
        if v == '({}, {})'.format(numerator, denominator):
            valid = True
            return k
    if not valid:
        raise ValueError('Invalid time signature: {}/{}'.format(numerator, denominator))
    

def convert_tempo_to_tempo_token(bpm):
    data = read_yaml('/home/longshen/work/MuseCoco/musecoco/midi_utils/tempo_dict.yaml')
    valid = False
    for k, v in data.items():
        v = v[1:-1].split(', ')
        if int(v[0]) <= bpm <= int(v[1]):
            valid = True
            return k
    if not valid:
        raise ValueError('Invalid tempo: {}'.format(bpm))
        

def tide_up_time_signature():
    ts = read_json('/home/longshen/work/MuseCoco/musecoco/midi_utils/ts_dict.json')
    ts_new = {'s-{}'.format(k): '({}, {})'.format(v[0], v[1]) for k, v in ts.items()}
    save_yaml(ts_new, '/home/longshen/work/MuseCoco/musecoco/midi_utils/ts_dict.yaml')


def generate_tempo_dict():
    tempo_dict = {}
    tempo_token_bpm_range = {}
    for bpm in range(10, 280):
        tok = 't-{}'.format(convert_tempo_to_id(bpm))
        if tok not in tempo_token_bpm_range:
            tempo_token_bpm_range[tok] = [999, -1]
        tempo_token_bpm_range[tok][0] = min(tempo_token_bpm_range[tok][0], bpm)
        tempo_token_bpm_range[tok][1] = max(tempo_token_bpm_range[tok][1], bpm)
        # tempo_dict[tok] = bpm
        # print(bpm, tok)
    # print(tempo_token_bpm_range)
    res = {k: '({}, {})'.format(v[0], v[1]) for k, v in tempo_token_bpm_range.items()}
    save_yaml(res, '/home/longshen/work/MuseCoco/musecoco/midi_utils/tempo_dict.yaml')

    # for i in range(0, ct, '/home/longshen/work/MuseCoco/musecoco/midi_utils/tempo_dict.yaml')



if __name__ == '__main__':
    main()


import os
import sys
dirof = os.path.dirname
sys.path.insert(0, dirof(__file__))

from legacy import remi_utils
from legacy import midiprocessor as mp
from legacy.midi_data_extractor.utils.pos_process import fill_pos_ts_and_tempo_
from legacy.midi_data_extractor.midi_processing import get_midi_pos_info, convert_pos_info_to_tokens
from legacy.midi_data_extractor.data_extractor import get_bar_positions, get_bars_insts, ChordDetector
from legacy.midiprocessor import MidiDecoder

ls = os.listdir
jpath = os.path.join


def _main():
    # test_remi_tok()
    check_chord_recognizer()
    # check_jingwei_chord_recognizer()


def test_remi_tok():
    dataset_dir = 'datasets/local/midis'
    output_dir = 'datasets/local/midis_recovered'
    remi_fp = 'remi_test.txt'
    midi_fns = ls(dataset_dir)
    midi_fns.sort()
    tk = RemiTokenizer()

    # remis = []
    # for midi_fn in midi_fns:
    #     midi_fp = jpath(dataset_dir, midi_fn)
    #     remi_toks = tk.midi_to_remi(midi_fp)
    #     remi_tok_strs = ' '.join(remi_toks)
    #     remis.append(remi_tok_strs + '\n')
    #
    # out_fp = remi_fp
    # with open(out_fp, 'w') as f:
    #     f.writelines(remis)

    with open(remi_fp) as f:
        remis = f.readlines()
    for id, remi in enumerate(remis):
        remi = remi.split(' ')
        output_fp = jpath(output_dir, 'midi_{}.mid'.format(id + 1))
        tk.remi_to_midi(remi, output_fp)


def check_chord_recognizer():
    '''
    Check Magenta's script to recognize chords from MIDI
    '''
    # midi_fp = './datasets/local/segmented/Track01876-3/midi.mid'
    midi_fp = './datasets/local/segmented/Track01501-1/midi.mid'
    tk = RemiTokenizer()
    tk.recognize_chord_from_midi(midi_fp)


# def check_jingwei_chord_recognizer():
#     from chord_recognition.main import transcribe_cb1000_midi
#     # midi_fp = './datasets/local/midis/test2.mid'
#     midi_fp = './datasets/local/segmented/Track01501-1/midi.mid'
#     transcribe_cb1000_midi(midi_fp, 'midi_chord.txt')


def _procedures():
    pass


def get_midi_length_in_seconds(midi_file_path):
    import mido

    mid = mido.MidiFile(midi_file_path)

    # 设置默认tempo（500000微秒每拍，即120拍每分钟）
    current_tempo = 500000
    midi_length = 0

    for track in mid.tracks:
        elapsed_time = 0
        for msg in track:
            # 如果是tempo事件，更新当前tempo
            if msg.type == 'set_tempo':
                current_tempo = msg.tempo
            elapsed_time += msg.time

        # 使用当前tempo将ticks时间转换为秒
        seconds = mido.tick2second(elapsed_time, mid.ticks_per_beat, current_tempo)
        midi_length = max(midi_length, seconds)

    return midi_length


class MidiUtil:
    @staticmethod
    def get_duration(midi_fp):
        midi_data = pretty_midi.PrettyMIDI(midi_fp)
        length_in_seconds = midi_data.get_end_time()
        return length_in_seconds


class RemiTokenizer:
    def __init__(self):
        self.encoding_method = "REMIGEN2"
        self.midi_encoder = mp.MidiEncoder(self.encoding_method)
        self.midi_decoder = MidiDecoder(self.encoding_method)
        self.chord_detector = ChordDetector(self.midi_encoder)

        self.attribute_list = ['I1s2', 'R1', 'R3', 'S2s1', 'S4', 'B1s1', 'TS1s1', 'K1', 'T1s1', 'P4', 'EM1', 'TM1']

    def midi_to_remi(self, 
                     midi_path, 
                     normalize_pitch, 
                     return_pitch_shift=False, 
                     return_key=False,
                     reorder_by_inst=False,
                     include_ts=True,
                     include_tempo=True,
                     include_velocity=True,
                     ):
        '''
        Convert a midi file to a remi sequence
        :param midi_path:
        :return: a list of str, remi token sequence.
        '''
        try:
            midi_obj = mp.midi_utils.load_midi(midi_path)
        except:
            print(f'Error: {midi_path}')
            return []
        pos_info = get_midi_pos_info(self.midi_encoder, midi_path=None, midi_obj=midi_obj, remove_empty_bars=False)
        # Longshen: don't remove empty bars to facilitate

        pos_info = fill_pos_ts_and_tempo_(pos_info)

        # Normalize pitch to C maj or A min.
        is_major = None
        normalize_pitch_value = normalize_pitch
        if normalize_pitch_value:
            try:
                pos_info, is_major, pitch_shift = self.midi_encoder.normalize_pitch(pos_info)
                # Pitch shift is an integer number which directly add towards the pitch of all notes inside pos_info
            except KeyboardInterrupt:
                raise
            except:
                is_major = None

        # Get bar token index, and instruments inside each bar
        bars_positions = get_bar_positions(pos_info)
        bars_instruments = get_bars_insts(pos_info, bars_positions)
        num_bars = len(bars_positions)
        assert num_bars == len(bars_instruments)

        # Get the tokens
        remi_seq = convert_pos_info_to_tokens(
            self.midi_encoder,
            pos_info,
            ignore_ts=False if include_ts is True else True,
            ignore_tempo=False if include_tempo is True else True,
            ignore_velocity=False if include_velocity is True else True,
        )
        assert remi_seq[-1] == 'b-1'

        # Re-organize the remi sequence by instrument
        if reorder_by_inst:
            bar_indices = remi_utils.from_remi_get_bar_idx(remi_seq)
            organized_remi = []
            for bar_id, (bar_start_idx, bar_end_idx) in bar_indices.items():
                bar_seq = remi_seq[bar_start_idx:bar_end_idx]
                new_bar = remi_utils.reorder_remi_bar(bar_seq, add_bar_token=True)
                organized_remi.extend(new_bar)
            remi_seq = organized_remi

        if return_pitch_shift:
            if return_key:
                return remi_seq, pitch_shift, is_major
            return remi_seq, pitch_shift
        else:
            return remi_seq

    def get_bar_pos(self, tokens):
        '''
        Get the index of starting tokens of each bar, from remi tokens.
        Specifically used for data preparation.

        :param tokens:
        :return:
        '''
        # Get the starting token of each bar
        start_token_index_of_the_bar = 0
        bar_id = 0
        bar_start_token_indices = {}
        # bars_token_positions[bar_id] = (start token index of this bar, start token index of next bar)
        for idx, token in enumerate(tokens):
            if token == 'b-1':
                start_token_index_of_next_bar = idx + 1
                bar_start_token_indices[bar_id] = (start_token_index_of_the_bar, start_token_index_of_next_bar)

                # Go to the next bar
                start_token_index_of_the_bar = start_token_index_of_next_bar
                bar_id = bar_id + 1
        return bar_start_token_indices

    def remi_to_midi(self, remi_seq, midi_path):
        '''
        Convert a list of remi tokens to a midi and save to disk.
        :param remi_seq: a list of str, each of which represent remi tokens.
        :param midi_path:
        :return:
        '''
        midi_obj = self.midi_decoder.decode_from_token_str_list(remi_seq)
        midi_obj.dump(midi_path)

    def remi_file_to_midi(self, remi_fp, midi_path):
        '''
        Convert a list of remi tokens to a midi and save to disk.
        :param remi_seq: a list of str, each of which represent remi tokens.
        :param midi_path:
        :return:
        '''
        with open(remi_fp) as f:
            t = f.read().strip()
        remi_seq = t.split(' ')

        # If have pitch token but no duration token,
        # Add duration token after each pitch
        has_pitch, has_duration = False, False
        for tok in remi_seq:
            if tok.startswith('p'):
                has_pitch = True
            if tok.startswith('d'):
                has_duration = True
        if has_pitch and not has_duration:
            new_remi_seq = []
            dur_token = 'd-6'
            for tok in remi_seq:
                if tok.startswith('p'):
                    new_remi_seq.append(tok)
                    new_remi_seq.append(dur_token)
                else:
                    new_remi_seq.append(tok)
            remi_seq = new_remi_seq

        midi_obj = self.midi_decoder.decode_from_token_str_list(remi_seq)
        midi_obj.dump(midi_path)

    def midi_to_remi_file(self, midi_path, remi_path, return_pitch_shift=False):
        if return_pitch_shift:
            remi_seq, pitch_shift = self.midi_to_remi(midi_path, return_pitch_shift)
        else:
            remi_seq = self.midi_to_remi(
                midi_path, 
                normalize_pitch=True,
                return_pitch_shift=False,
                return_key=False,
                reorder_by_inst=True,
                include_ts=False,
                include_tempo=False,
                include_velocity=False,
            )
            pitch_shift = None
        remi_tok_strs = ' '.join(remi_seq)
        with open(remi_path, 'w') as f:
            f.write(remi_tok_strs + '\n')

        if return_pitch_shift:
            return pitch_shift

    def recognize_chord_from_midi_old(self, midi_path):
        '''
        Method by MuseCoco. Not recommended due to inaccurate detection.

        Recognize chord info from midi file using Magenta's script
        :param midi_path: path of midi file.
        :return: a list of string, recognized chords
        '''
        midi_obj = mp.midi_utils.load_midi(midi_path)
        pos_info = get_midi_pos_info(self.midi_encoder, midi_path=None, midi_obj=midi_obj, remove_empty_bars=False)
        # Longshen: don't remove empty bars to facilitate
        pos_info = fill_pos_ts_and_tempo_(pos_info)
        # pos_info, is_major, _ = self.midi_encoder.normalize_pitch(pos_info)
        bars_chords = self.chord_detector.infer_chord_for_pos_info(pos_info)
        # print(bars_chords)
        # exit(10)
        return bars_chords


class RemiUtil:
    def __init__(self):
        pass

    @staticmethod
    def extract_condition_from_remi(remi_seq, key, dir):
        '''
        Dir: path to save the segment midis
        '''
        bar_indices = RemiUtil.get_bar_idx_from_remi(remi_seq)
        num_bars = len(bar_indices)
        tk = RemiTokenizer()

        # print(num_bars)
        if num_bars == 0:
            raise Exception('Bar num = 0')

        all_bar_info = {}
        for bar_id in bar_indices:
            bar_start_idx, bar_end_idx = bar_indices[bar_id]
            bar_remi = remi_seq[bar_start_idx:bar_end_idx]
            bar_info = {}
            bar_info['remi_seq'] = bar_remi

            # Detokenize the segment
            midi_seg_fp = jpath(dir, f'bar_{bar_id}.mid')
            tk.remi_to_midi(bar_remi, midi_seg_fp)

            # Key and scale
            key_segment = key
            bar_info['is_major'] = key_segment
            bar_info['scale'] = 1

            # Instruments
            insts_this_bar = set()
            for token in bar_remi:
                if token.startswith('i-'):
                    insts_this_bar.add(token)
            insts_this_bar = list(insts_this_bar)
            insts_this_bar = sorted(insts_this_bar, key=lambda x: int(x.split('-')[1]))  # sort by inst id
            bar_info['instruments'] = insts_this_bar

            # Recognize chord
            chords = recognize_chord_from_midi(midi_seg_fp, out_fp=None)
            dur = MidiUtil.get_duration(midi_seg_fp)
            chord_quantized = quantize_chord(chords, dur=dur, num_bars=1, num_chord_per_bar=2).tolist()
            bar_info['chords'] = chord_quantized

            # Obtain time function
            txt_seq = get_time_function_from_remi_one_bar(bar_remi)
            bar_info['onset_density'] = txt_seq

            all_bar_info[bar_id] = bar_info

        return all_bar_info
    
    @staticmethod
    def extract_condition_for_ss_from_remi(remi_seq, key, dir):
        '''
        Dir: path to save the segment midis

        Extract the condition info for each bar in the remi
        '''
        bar_indices = RemiUtil.get_bar_idx_from_remi(remi_seq)
        num_bars = len(bar_indices)
        tk = RemiTokenizer()

        # print(num_bars)
        if num_bars == 0:
            raise Exception('Bar num = 0')

        all_bar_info = {}
        for bar_id in bar_indices:
            bar_start_idx, bar_end_idx = bar_indices[bar_id]
            bar_remi = remi_seq[bar_start_idx:bar_end_idx]
            bar_info = {}
            bar_info['remi_seq'] = bar_remi

            input_tokens = []
            ''' Only retain position, pitch, and bar line '''
            bar_inp_seq = []
            bar_remi_seq = bar_remi
            for tok in bar_remi_seq:
                if tok.startswith('o-') or tok.startswith('p-') or tok.startswith('b-1'):
                    bar_inp_seq.append(tok)
            input_tokens.extend(bar_inp_seq)

            ''' Add Inst info '''
            # Obtain instruments from bar_seq
            insts_this_bar = set()
            for token in bar_remi_seq:
                if token.startswith('i-'):
                    insts_this_bar.add(token)
            insts_this_bar = list(insts_this_bar)
            insts_this_bar = sorted(insts_this_bar, key=lambda x: int(x.split('-')[1]))  # sort by inst id
            # Add instrument info to input_tokens
            input_tokens.extend(insts_this_bar)
            input_tokens.append('INS')

            bar_info['input_tokens'] = input_tokens

            # # Detokenize the segment
            # midi_seg_fp = jpath(dir, f'bar_{bar_id}.mid')
            # tk.remi_to_midi(bar_remi, midi_seg_fp)

            # # Key and scale
            # key_segment = key
            # bar_info['is_major'] = key_segment
            # bar_info['scale'] = 1

            # # Instruments
            # insts_this_bar = set()
            # for token in bar_remi:
            #     if token.startswith('i-'):
            #         insts_this_bar.add(token)
            # insts_this_bar = list(insts_this_bar)
            # insts_this_bar = sorted(insts_this_bar, key=lambda x: int(x.split('-')[1]))  # sort by inst id
            # bar_info['instruments'] = insts_this_bar

            # # Recognize chord
            # chords = recognize_chord_from_midi(midi_seg_fp, out_fp=None)
            # dur = MidiUtil.get_duration(midi_seg_fp)
            # chord_quantized = quantize_chord(chords, dur=dur, num_bars=1, num_chord_per_bar=2).tolist()
            # bar_info['chords'] = chord_quantized

            # # Obtain time function
            # txt_seq = get_time_function_from_remi_one_bar(bar_remi)
            # bar_info['onset_density'] = txt_seq

            all_bar_info[bar_id] = bar_info

        return all_bar_info
    
    @staticmethod
    def extract_condition_for_sss_ipo_from_remi_segment(remi_seq):
        '''
        Extract the condition info for each segment in the remi
        '''
        bar_indices = RemiUtil.get_bar_idx_from_remi(remi_seq)
        num_bars = len(bar_indices)
        tk = RemiTokenizer()

        # print(num_bars)
        if num_bars == 0:
            raise Exception('Bar num = 0')
        
        # Obtain the indices of each segment
        segment_bar_num = 2
        segment_indices = []
        hop_bar = 1
        for start_bar_id in range(0, num_bars - segment_bar_num + 1, hop_bar):  # 1 000 000
            bar_ids = [start_bar_id + i for i in range(segment_bar_num)]
            start_bar_id = bar_ids[0]
            end_bar_id = bar_ids[-1]
            segment_start_idx = bar_indices[start_bar_id][0]
            segment_end_idx = bar_indices[end_bar_id][1]
            segment_indices.append((segment_start_idx, segment_end_idx))

        all_seg_info = []
        for seg_start_idx, seg_end_idx in segment_indices:
            seg_remi = remi_seq[seg_start_idx:seg_end_idx]
            seg_info = {}
            seg_info['remi_seq'] = seg_remi

            # Obtain input tokens from segment remi
            from dataset_preparation.create_midi_only_dataset import TwoBarDatasetPreparation
            preprocess = TwoBarDatasetPreparation()
            input_tokens = preprocess.obtain_input_tokens_from_remi_seq_sss_ipo(seg_remi)
            seg_info['input_tokens'] = input_tokens

            all_seg_info.append(seg_info)

        return all_seg_info
    
    @staticmethod
    def extract_condition_for_sss_ipo_ps_from_remi_segment(remi_seq):
        '''
        Extract the condition info for each segment in the remi
        '''
        bar_indices = RemiUtil.get_bar_idx_from_remi(remi_seq)
        num_bars = len(bar_indices)
        tk = RemiTokenizer()

        # print(num_bars)
        if num_bars == 0:
            raise Exception('Bar num = 0')
        
        # Obtain the indices of each segment
        segment_bar_num = 2
        segment_indices = []
        hop_bar = 1
        for start_bar_id in range(0, num_bars - segment_bar_num + 1, hop_bar):  # 1 000 000
            bar_ids = [start_bar_id + i for i in range(segment_bar_num)]
            start_bar_id = bar_ids[0]
            end_bar_id = bar_ids[-1]
            segment_start_idx = bar_indices[start_bar_id][0]
            segment_end_idx = bar_indices[end_bar_id][1]
            segment_indices.append((segment_start_idx, segment_end_idx))

        all_seg_info = []
        for seg_start_idx, seg_end_idx in segment_indices:
            seg_remi = remi_seq[seg_start_idx:seg_end_idx]
            seg_info = {}
            seg_info['remi_seq'] = seg_remi

            # Obtain input tokens from segment remi
            from dataset_preparation.create_midi_only_dataset import TwoBarDatasetPreparation
            preprocess = TwoBarDatasetPreparation()
            input_tokens = preprocess.obtain_input_tokens_from_remi_seq_sss_ipo_sort_pitch(seg_remi)
            seg_info['input_tokens'] = input_tokens

            all_seg_info.append(seg_info)

        return all_seg_info
    
    @staticmethod
    def extract_condition_for_sss_ipo_tf_from_remi_segment(remi_seq):
        '''
        Extract the condition info for each segment in the remi
        '''
        bar_indices = RemiUtil.get_bar_idx_from_remi(remi_seq)
        num_bars = len(bar_indices)
        tk = RemiTokenizer()

        # print(num_bars)
        if num_bars == 0:
            raise Exception('Bar num = 0')
        
        # Obtain the indices of each segment
        segment_bar_num = 2
        segment_indices = []
        hop_bar = 1
        for start_bar_id in range(0, num_bars - segment_bar_num + 1, hop_bar):  # 1 000 000
            bar_ids = [start_bar_id + i for i in range(segment_bar_num)]
            start_bar_id = bar_ids[0]
            end_bar_id = bar_ids[-1]
            segment_start_idx = bar_indices[start_bar_id][0]
            segment_end_idx = bar_indices[end_bar_id][1]
            segment_indices.append((segment_start_idx, segment_end_idx))

        all_seg_info = []
        for seg_start_idx, seg_end_idx in segment_indices:
            seg_remi = remi_seq[seg_start_idx:seg_end_idx]
            seg_info = {}
            seg_info['remi_seq'] = seg_remi

            # Obtain input tokens from segment remi
            from dataset_preparation.create_midi_only_dataset import TwoBarDatasetPreparation
            preprocess = TwoBarDatasetPreparation()
            input_tokens = preprocess.obtain_input_tokens_from_remi_sss_ipo_tf(seg_remi)
            seg_info['input_tokens'] = input_tokens

            all_seg_info.append(seg_info)

        return all_seg_info

    @staticmethod
    def get_bar_idx_from_remi(remi_seq):
        # Get the starting token of each bar
        start_token_index_of_the_bar = 0
        bar_id = 0
        bar_indices = {}

        # bars_token_positions[bar_id] = (start token index of this bar, start token index of next bar)
        for idx, token in enumerate(remi_seq):
            if token == 'b-1':
                start_token_index_of_next_bar = idx + 1
                bar_indices[bar_id] = (start_token_index_of_the_bar, start_token_index_of_next_bar)

                # Go to the next bar
                start_token_index_of_the_bar = start_token_index_of_next_bar
                bar_id = bar_id + 1
        return bar_indices

    @staticmethod
    def get_inst_from_remi(remi_seq):
        '''
        Get a list of instrument tokens from a sequence of remi tokens, for each bar of info
        There is no repeated instrument in the list.
        Each list is sorted by program id.
        e.g., [
                [i-0, i-8, i-128], # insts of 1st bar
                [i-0, i-128],    # insts of 2nd bar
            ]
        '''
        b_1_indices = [index for index, element in enumerate(remi_seq) if element == "b-1"]
        num_bars = len(b_1_indices)
        b_1_indices.insert(0, -1)

        insts_each_bar = []
        for bar_id in range(num_bars):
            bar_start_idx = b_1_indices[bar_id] + 1
            bar_end_idx = b_1_indices[bar_id + 1]
            bar_seq = remi_seq[bar_start_idx:bar_end_idx]

            # Obtain instruments from bar_seq
            insts_this_bar = set()
            for token in bar_seq:
                if token.startswith('i-'):
                    insts_this_bar.add(token)
            insts_this_bar = list(insts_this_bar)
            insts_this_bar = sorted(insts_this_bar, key=lambda x: int(x.split('-')[1]))  # sort by inst id

            # Add instrument info to input_tokens
            insts_each_bar.append(insts_this_bar)

        return insts_each_bar

    @staticmethod
    def get_inst_from_input_seq(inp_seq):
        input_list = inp_seq

        # 初始化结果列表和临时子列表
        result = []
        current_sublist = []

        for item in input_list:
            if item.startswith('i-'):
                # 如果遇到以'i-'开头的元素，将它加入sub list
                current_sublist.append(item)
            else:  # 如果元素不以i开头，那么将sub list加入result
                if len(current_sublist) > 0:
                    result.append(current_sublist)
                    current_sublist = []

        # 循环结束后，如果当前子列表非空，将其添加到结果中
        if len(current_sublist) > 0:
            result.append(current_sublist)

        return result

    @staticmethod
    def get_chord_from_input_seq(input_seq):
        '''
        Get chord info from input sequence
        CR-X, CT-Y
        :param input_seq:
        :return: Two lists, first is chord root, second is chord type.
        e.g., [[CR-0, CR-2], [CR-1, CR-1]],    [[CT-5, CT-7], [CT-5, CT-0]]
        '''
        # 初始化结果列表和临时子列表
        res_cr = []
        res_ct = []

        # chord_root_cur_bar = []
        # chord_type_cur_bar = []
        for token in input_seq:
            if token.startswith('CR'):
                # 如果遇到以'i-'开头的元素，将它加入sub list
                # chord_root_cur_bar.append(token)
                res_cr.append(token)
            elif token.startswith('CT'):
                # chord_type_cur_bar.append(token)
                res_ct.append(token)
            # elif token == 'b-1':
            #     res_cr.append(chord_root_cur_bar)
            #     res_ct.append(chord_type_cur_bar)
            #     chord_root_cur_bar = []
            #     chord_type_cur_bar = []

        return res_cr, res_ct

    @staticmethod
    def get_rhythm_from_input_seq(inp_seq):
        '''
        Get the onset density on each position, from the input sequence.
        :param inp_seq: input condition token sequence. May contain info for several bars.
        :return: a list of dict, each dict is the rhythm info inside a bar. Key is position, value is onset count.
            [
                {
                    2: 8, 3, 4,
                },
                {
                    0: 10, 8: 5,
                }
            ]
        '''
        res = []
        for tok in inp_seq:
            if tok == 'TF':
                t = {}
            elif tok == 'b-1':
                res.append(t)
            elif tok.startswith('o-'):
                cur_pos = int(tok.split('-')[-1])
            elif tok.startswith('TF-'):
                onset_cnt = int(tok.split('-')[-1])
                t[cur_pos] = onset_cnt
        return res

    @staticmethod
    def read_remi(fp, split=True, remove_input=False):
        with open(fp) as f:
            remi_str = f.readline().strip()

        if remove_input:
            remi_str = remi_str.split(' <sep> ')[1]

        if split:
            remi_seq = remi_str.split(' ')
        else:
            remi_seq = remi_str
        return remi_seq

    @staticmethod
    def save_remi_seq(remi_seq, fp):
        with open(fp, 'w') as f:
            f.write(' '.join(remi_seq))


if __name__ == '__main__':
    _main()



import os
import sys
dirof = os.path.dirname
sys.path.insert(0, dirof(__file__))

import miditoolkit
import pretty_midi
from utils import read_yaml
from typing import List, Dict, Tuple
from midi_encoding import MidiEncoder, load_midi, fill_pos_ts_and_tempo_, convert_tempo_to_id, convert_id_to_tempo
from time_signature_utils import TimeSignatureUtil
from keys_normalization import detect_key

class Note:
    def __init__(self, onset:int, duration:int, pitch:int, velocity:int=64, is_drum=False):
        '''
        Create an instance of a Note object.

        Parameters
        ----------
        onset : int
            The onset time of the note.
            NOTE: The time unit is "position", which is 48th note (1/12 of a beat). Usually, one bar has 48 positions.
            Range: [0, 127] The maximum allowed onset position is 127, e.g., at the end of 127*1/12 = 10.58 beats.
        duration : int
            The duration of the note.
            NOTE: The time unit is "position", which is 48th note (1/12 of a beat).
            And, the duration of a note should be greater than 1 (smaller value will be rounded to 1).
            The maximum allowed duration is 127, e.g., has length of 127*1/12 = 10.58 beats.
        pitch : int
            The MIDI pitch of the note.
            The pitch value should be in the range of [0, 255].
                0~127: MIDI pitch
        velocity : int
            The velocity of the note.
        '''
        assert isinstance(onset, int), "onset must be an integer"
        assert isinstance(duration, int), "duration must be an integer"
        assert isinstance(pitch, int), "pitch must be an integer"
        assert isinstance(velocity, int), "velocity must be an integer"
        assert 0 <= onset <= 127, "onset must be in the range of [0, 127]"
        assert 0 <= pitch <= 255, "pitch must be in the range of [0, 255]"
        assert 0 <= velocity <= 127, "velocity must be in the range of [0, 127]"
        # if is_drum:
        #     assert 128 <= pitch <= 255, "Drum pitch must be in the range of [128, 255]"
        # else:
        assert 0 <= pitch <= 127, "MIDI pitch must be in the range of [0, 127]"

        # Round the values
        duration = min(max(1, duration), 127) # duration must be in the range of [1, 127]

        self.onset = onset
        self.duration = duration
        self.pitch = pitch
        self.velocity = velocity

    def __str__(self) -> str:
        return f'(o:{self.onset},d:{self.duration},p:{self.pitch},v:{self.velocity})'
    
    def __repr__(self) -> str:
        return self.__str__()

    def __lt__(self, other):
        if self.onset != other.onset:
            return self.onset < other.onset
        else:
            return self.pitch > other.pitch


class Track:
    '''
    This class save all notes for a same track within a bar.
    '''
    def __init__(self, inst_id, notes:Dict[int, List[Note]]):
        self.inst_id = inst_id
        self.non_empty_pos = list(notes.keys())
        self.non_empty_pos.sort()

        if inst_id == 128:
            self.is_drum = True
        else:
            self.is_drum = False

        self.notes = []
        for pos, notes_of_pos in notes.items():
            for note in notes_of_pos:
                pitch, duration, velocity = note
                note_instance = Note(onset=pos, duration=duration, pitch=pitch, velocity=velocity, is_drum=self.is_drum)
                self.notes.append(note_instance)
        self.notes.sort()

        # Calculate the average pitch
        if self.is_drum:
            self.avg_pitch = -1
        else:
            pitches = [note.pitch for note in self.notes]
            self.avg_pitch = sum(pitches) / len(pitches)

    def __str__(self) -> str:
        return f'Inst {self.inst_id}: {len(self.notes)} notes'
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def __len__(self):
        return len(self.notes)
    
    def __lt__(self, other):
        '''
        Track with higher pitch will be placed at the front (more important)
        '''
        return self.avg_pitch > other.avg_pitch
    
    def get_note_list(self) -> List[Tuple[int, int, int, int]]:
        '''
        Get all notes in the Track.

        Returns:
            List of tuples (onset, pitch, duration, velocity)
        '''
        all_notes = []
        for note in self.notes:
            all_notes.append((note.onset, note.pitch, note.duration, note.velocity))
        return all_notes


class Bar:
    def __init__(self, id, notes_of_insts:Dict[int, Dict[int, List]], time_signature=None, tempo=None):
        '''
        NOTE: The instrument with higher average pitch will be placed at the front.
        '''
        if time_signature:
            assert isinstance(time_signature, tuple), "time_signature must be a tuple"
        else:
            time_signature = (4, 4)
        if tempo:
            assert isinstance(tempo, (int, float)), "tempo must be an integer or float"
        else:
            tempo = 120.0

        # Round tempo to 0.01
        tempo = round(tempo, 2)

        self.bar_id = id
        track_list = []
        self.tracks: Dict[int, Track] = {}
        for inst_id, notes in notes_of_insts.items():
            track = Track(inst_id, notes)
            track_list.append(track)
        track_list.sort()
        for track in track_list:
            inst_id = track.inst_id
            self.tracks[inst_id] = track

        self.time_signature = time_signature
        self.tempo = tempo

    def __len__(self):
        return len(self.tracks)
    
    def __str__(self) -> str:
        return f'Bar {self.bar_id}: {len(self.tracks)} insts'

    def __repr__(self) -> str:
        return self.__str__()
    

class MultiTrack:
    def __init__(self, bars:List[Bar]):
        '''
        Args:
            bars: List of Bar objects
            pitch_shift: The pitch shift value. None means not detected.
            is_major: The major/minor key information. None means not detected.
        '''
        self.bars = bars 

        # Load the time signature dictionary
        ts_fp = os.path.join(os.path.dirname(__file__), 'dict_time_signature.yaml')
        self.ts_dict = read_yaml(ts_fp)

    def __len__(self):
        return len(self.bars)
    
    def __getitem__(self, idx):
        if isinstance(idx, int):
            bar_subset = [self.bars[idx]]
            return self.bars[idx]
        elif isinstance(idx, slice):
            bar_subset = self.bars[idx]
            return MultiTrack(bars=bar_subset)
    
    def __str__(self) -> str:
        return f'MultiTrack: {len(self.bars)} bars'
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def normalize_pitch(self):
        '''
        Normalize the pitch of all notes in the MultiTrack object.
        '''

        ''' Detect major/minor key and pitch shift needed for the key normalization '''
        is_major, pitch_shift = self.detect_key()

        ''' Apply the pitch shift to the notes '''
        for bar in self.bars:
            for inst_id, track in bar.tracks.items():
                if track.is_drum:
                    continue
                else:
                    for note in track.notes:
                        note.pitch += pitch_shift
        return self
    
    def detect_key(self):
        '''
        Determine the major/minor key and pitch shift needed for key normalization.

        Returns:
        - is_major: True if major, False if minor
        - pitch_shift: The pitch shift needed for key normalization
        '''
        note_list = self.get_note_list(with_drum=False)
        is_major, pitch_shift = detect_key(note_list)
        return is_major, pitch_shift

    def get_unique_insts(self):
        '''
        Get all unique instruments in the MultiTrack object.
        '''
        all_insts = set()
        for bar in self.bars:
            for inst_id in bar.tracks.keys():
                all_insts.add(inst_id)
        return all_insts

    @classmethod
    def from_midi(cls, midi_fp:str):
        '''
        Create a MultiTrack object from a MIDI file.
        '''
        assert isinstance(midi_fp, str), "midi_fp must be a string"
        assert os.path.exists(midi_fp), "midi_fp does not exist"

        midi_encoder = MidiEncoder()
        midi_obj = load_midi(midi_fp)

        # Obtain information for each position (a dense representation)
        pos_info = midi_encoder.collect_pos_info(
            midi_obj, 
            trunc_pos=None, 
            tracks=None, 
            remove_same_notes=False, 
            end_offset=0
        )
        # bar: every pos
        # ts: only at pos where it changes, otherwise None
        # in-bar position
        # tempo: only at pos where it changes, otherwise None
        # insts_notes: only at pos where the note starts, otherwise None

        # Fill time signature and tempo info to the first pos of each bar
        pos_info = fill_pos_ts_and_tempo_(pos_info)

        # # Determine pitch normalization and major/minor info
        # _, is_major, pitch_shift = midi_encoder.normalize_pitch(pos_info) # Can make error some times. Not sure about direction of pitch shift yet.
        # # If apply this pitch shift to the original MIDI, the key will be C major or A minor

        # Generate bar sequences and note sequences
        bar_seqs = []
        bar_id_prev_pos = -1
        cur_ts = None
        cur_tempo = None
        for i in range(len(pos_info)):
            bar_id, ts, pos, tempo, insts_notes = pos_info[i]

            # Update time signature and tempo
            if ts is not None:
                cur_ts = ts
            if tempo is not None:
                cur_tempo = tempo
            
            # Determine if this is a new bar
            if bar_id > bar_id_prev_pos:
                notes_of_instruments = {}

            # Add the note info
            if insts_notes is not None:
                for inst_id, notes in insts_notes.items():
                    # each note contain [pitch, duration, velocity]
                    if inst_id not in notes_of_instruments:
                        notes_of_instruments[inst_id] = {}
                    if pos not in notes_of_instruments[inst_id]:
                        notes_of_instruments[inst_id][pos] = []
                    notes_of_instruments[inst_id][pos].extend(notes)

            # Determine if this is the last position of a bar
            last_pos_of_bar = False
            if i == len(pos_info) - 1:
                last_pos_of_bar = True
            else:
                next_bar_id, _, next_pos, _, _ = pos_info[i+1]
                if next_bar_id != bar_id:
                    last_pos_of_bar = True
            
            # Add the bar info
            if last_pos_of_bar:
                bar_instance = Bar(
                    id=bar_id, 
                    notes_of_insts=notes_of_instruments,
                    time_signature=cur_ts,
                    tempo=cur_tempo,
                )
                bar_seqs.append(bar_instance)
                
            bar_id_prev_pos = bar_id

        return cls(bars=bar_seqs)
    
    @classmethod
    def from_remiz_seq(cls, remiz_seq:List[str]):
        '''
        Create a MultiTrack object from a remiz sequence.
        '''
        assert isinstance(remiz_seq, list), "remiz_seq must be a list"

        remiz_str = ' '.join(remiz_seq)
        return cls.from_remiz_str(remiz_str)

    @classmethod
    def from_remiz_str(cls, remiz_str:str):
        '''
        Create a MultiTrack object from a remiz string.
        '''
        assert isinstance(remiz_str, str), "remiz_str must be a string"
        assert 'b-1' in remiz_str, "remiz_str must contain at least one bar"
        if 'v' in remiz_str:
            with_velocity = True
        else:
            with_velocity = False

        bar_seqs = []

        # Split to bars
        bar_strs = remiz_str.split('b-1')
        bar_strs.pop()
        for bar_id, bar_str in enumerate(bar_strs):
            bar_seq = bar_str.strip().split()
            
            time_sig = None
            tempo = None
            need_create_note = False
            notes_of_instruments = {}
            for tok in bar_seq:
                if tok.startswith('s-'):
                    time_sig = TimeSignatureUtil.convert_time_signature_token_to_tuple(tok)
                elif tok.startswith('t-'):
                    tempo_id = int(tok[2:])
                    tempo = convert_id_to_tempo(tempo_id)
                elif tok.startswith('i-'):
                    inst_id = int(tok[2:])
                elif tok.startswith('o-'):
                    pos = int(tok[2:])
                elif tok.startswith('p-'):
                    pitch = int(tok[2:])
                    if pitch >= 128:
                        pitch -= 128
                elif tok.startswith('d-'):
                    duration = int(tok[2:])
                    if not with_velocity:
                        velocity = 64
                        need_create_note = True
                elif tok.startswith('v-'):
                    velocity = int(tok[2:])
                    need_create_note = True

                if need_create_note:
                    if inst_id not in notes_of_instruments:
                        notes_of_instruments[inst_id] = {}
                    if pos not in notes_of_instruments[inst_id]:
                        notes_of_instruments[inst_id][pos] = []
                    notes_of_instruments[inst_id][pos].append([pitch, duration, velocity])
                    need_create_note = False
            
            # Create a Bar instance
            bar_instance = Bar(
                id=bar_id, 
                notes_of_insts=notes_of_instruments,
                time_signature=time_sig,
                tempo=tempo,
            )
            bar_seqs.append(bar_instance)
                
        return cls(bars=bar_seqs)

    def to_remiz_seq(self, with_ts=False, with_tempo=False, with_velocity=False):
        '''
        Convert the MultiTrack object to a REMI-z sequence of tokens.
        '''
        ret = []
        for bar in self.bars:
            bar_seq = []

            # Add time signature
            if with_ts:
                # time_sig = bar.time_signature.strip()[1:-1]
                num, den = bar.time_signature
                ts_token = self.convert_time_signature_to_ts_token(int(num), int(den))
                bar_seq.append(ts_token)

            if with_tempo:
                tempo_id = convert_tempo_to_id(bar.tempo)
                tempo_tok = f't-{tempo_id}'
                bar_seq.append(tempo_tok)
                
            for inst_id, track in bar.tracks.items():
                track_seq = [f'i-{inst_id}']
                prev_pos = -1
                for note in track.notes:
                    if note.onset > prev_pos:
                        track_seq.append(f'o-{note.onset}')
                        prev_pos = note.onset

                    if track.is_drum:
                        pitch_id = note.pitch + 128
                    else:
                        pitch_id = note.pitch
                    track_seq.extend([
                        f'p-{pitch_id}',
                        f'd-{note.duration}',
                    ])

                    if with_velocity:
                        track_seq.append(f'v-{note.velocity}')
                bar_seq.extend(track_seq)
            bar_seq.append('b-1')
            ret.extend(bar_seq) 
        return ret
    
    def to_remiz_str(self, with_ts=False, with_tempo=False, with_velocity=False):
        '''
        Convert the MultiTrack object to a REMI-z string.
        '''
        ret = self.to_remiz_seq(with_ts=with_ts, with_tempo=with_tempo, with_velocity=with_velocity)
        return ' '.join(ret)

    def to_midi(self, midi_fp: str):
        """
        Create a MIDI file from the MultiTrack object using miditoolkit.
        """
        assert isinstance(midi_fp, str), "midi_fp must be a string"
        
        # 创建一个空的 MidiFile 对象
        # 默认 ticks_per_beat 是480，你可以根据需要修改
        midi_obj = miditoolkit.midi.parser.MidiFile(ticks_per_beat=480)
        ticks_per_beat = midi_obj.ticks_per_beat

        # 如果有小节，则获取初始速度，否则用默认值120 BPM
        if len(self.bars) > 0:
            initial_tempo = self.bars[0].tempo
        else:
            initial_tempo = 120.0

        # 初始化时间线计数（以ticks为单位）
        cumulative_bar_ticks = 0

        # 插入初始速度与拍号（如果有小节）
        if len(self.bars) > 0:
            # 初始拍号
            numerator, denominator = self.bars[0].time_signature
            # TimeSignature 与 TempoChange 用 ticks 来定位事件位置
            midi_obj.time_signature_changes.append(
                miditoolkit.midi.containers.TimeSignature(
                    numerator=numerator,
                    denominator=denominator,
                    time=0  # 第一个小节拍号从0 tick开始
                )
            )
            
            # 初始速度
            midi_obj.tempo_changes.append(
                miditoolkit.midi.containers.TempoChange(
                    tempo=initial_tempo,
                    time=0  # 初始速度从0 tick开始生效
                )
            )

        # 乐器映射表：inst_id -> Instrument对象
        instrument_map = {}

        last_time_signature = self.bars[0].time_signature if len(self.bars) > 0 else None
        last_tempo = initial_tempo

        # 遍历每个小节
        for bar_index, bar in enumerate(self.bars):
            # 如果拍号变了，插入新的 TimeSignature 事件
            if bar.time_signature != last_time_signature:
                numerator, denominator = bar.time_signature
                midi_obj.time_signature_changes.append(
                    miditoolkit.midi.containers.TimeSignature(
                        numerator=numerator,
                        denominator=denominator,
                        time=cumulative_bar_ticks
                    )
                )
                last_time_signature = bar.time_signature

            # 如果速度变了，插入新的 TempoChange 事件
            if bar.tempo != last_tempo:
                midi_obj.tempo_changes.append(
                    miditoolkit.midi.containers.TempoChange(
                        tempo=bar.tempo,
                        time=cumulative_bar_ticks
                    )
                )
                last_tempo = bar.tempo

            # 计算本小节的长度（以拍为单位）
            # 拍数 = 分子 * (4 / 分母)
            # 与之前一样的计算方法
            beats_per_bar = bar.time_signature[0] * (4.0 / bar.time_signature[1])
            # 小节长度(以ticks为单位)
            bar_length_ticks = int(beats_per_bar * ticks_per_beat)

            # 为当前小节中的音符计算相对时间（转换为ticks）
            for inst_id, track in bar.tracks.items():
                # 获取或创建Instrument
                if inst_id not in instrument_map:
                    program = 0 if inst_id == 128 else inst_id
                    # 创建乐器（Instrument）
                    # miditoolkit不强制要求不同instrument_id映射到特定音色，你可以根据实际需要调整program值。
                    instrument = miditoolkit.midi.containers.Instrument(
                        program=program,
                        is_drum=(inst_id == 128),  # 若为打击乐
                        name=f"Instrument_{inst_id}"
                    )
                    instrument_map[inst_id] = instrument
                    midi_obj.instruments.append(instrument)
                else:
                    instrument = instrument_map[inst_id]

                for note in track.notes:
                    # onset和duration是以某种beats为单位（如之前为12分音符换算）
                    onset_time_beats = note.onset / 12.0
                    duration_beats = note.duration / 12.0

                    # 转换为ticks
                    note_start = cumulative_bar_ticks + int(onset_time_beats * ticks_per_beat)
                    note_end = note_start + int(duration_beats * ticks_per_beat)

                    midi_note = miditoolkit.midi.containers.Note(
                        velocity=note.velocity,
                        pitch=note.pitch,
                        start=note_start,
                        end=note_end
                    )
                    instrument.notes.append(midi_note)

            # 更新累计的ticks数，以便下一个小节从正确的时间点开始
            cumulative_bar_ticks += bar_length_ticks

        # 写入MIDI文件
        midi_obj.dump(midi_fp)
        print(f"MIDI file successfully written to {midi_fp}")


    def to_midi_prettymidi(self, midi_fp: str):
        """
        Create a MIDI file from the MultiTrack object.

        Deprecated: Use to_midi() instead. Because this version cannot handle tempo changes.
        """
        assert isinstance(midi_fp, str), "midi_fp must be a string"
        
        import numpy as np
        import pretty_midi
        
        

        # Initialize instrument map
        instrument_map = {}

        # Track the cumulative time for each bar
        cumulative_bar_time = 0.0
        last_time_signature = None

        # Initialize arrays for tempo changes
        tempo_change_times = []
        tempi = []

        # Get initial tempo from the first bar, if available
        if len(self.bars) > 0:
            initial_tempo = self.bars[0].tempo
        else:
            # If no bars, fallback to a default tempo (e.g. 120 bpm)
            initial_tempo = 120.0

        # Set the initial tempo event at time zero
        tempo_change_times.append(0.0)
        tempi.append(60_000_000 / initial_tempo)
        last_tempo = initial_tempo

        # Create a PrettyMIDI object
        midi = pretty_midi.PrettyMIDI(
            initial_tempo=initial_tempo
        )

        for bar in self.bars:
            # Handle time signature changes
            if bar.time_signature != last_time_signature:
                numerator, denominator = bar.time_signature
                midi.time_signature_changes.append(
                    pretty_midi.TimeSignature(numerator, denominator, cumulative_bar_time)
                )
                last_time_signature = bar.time_signature

            # Handle tempo changes
            if bar.tempo != last_tempo:
                tempo_in_microseconds = 60_000_000 / bar.tempo
                tempo_change_times.append(cumulative_bar_time)
                tempi.append(tempo_in_microseconds)
                last_tempo = bar.tempo

            # Iterate over tracks in the current bar
            for inst_id, track in bar.tracks.items():
                # Ensure instrument_map is used correctly
                if inst_id not in instrument_map:
                    program = 0 if inst_id == 128 else inst_id
                    instrument = pretty_midi.Instrument(program=program, is_drum=(inst_id == 128))
                    instrument_map[inst_id] = instrument
                    midi.instruments.append(instrument)
                else:
                    instrument = instrument_map[inst_id]

                for note in track.notes:
                    onset_time_beats = note.onset / 12.0
                    onset_time_seconds = cumulative_bar_time + (onset_time_beats * (60.0 / bar.tempo))
                    duration_beats = note.duration / 12.0
                    duration_seconds = duration_beats * (60.0 / bar.tempo)

                    midi_note = pretty_midi.Note(
                        velocity=note.velocity,
                        pitch=note.pitch,
                        start=onset_time_seconds,
                        end=onset_time_seconds + duration_seconds
                    )
                    instrument.notes.append(midi_note)

            # Calculate beats per bar using the current bar's time signature
            beats_per_bar = bar.time_signature[0] * (4 / bar.time_signature[1])  # Numerator * (4 / Denominator)
            bar_duration = beats_per_bar * (60.0 / bar.tempo)
            cumulative_bar_time += bar_duration

        # Set tempo changes in PrettyMIDI
        midi._tempo_change_times = np.array(tempo_change_times)
        midi._tempi = np.array(tempi)

        # Write MIDI file
        midi.write(midi_fp)
        print(f"MIDI file successfully written to {midi_fp}")

    def convert_time_signature_to_ts_token(self, numerator, denominator):
        ts_dict = self.ts_dict
        valid = False
        for k, v in ts_dict.items():
            if v == '({}, {})'.format(numerator, denominator):
                valid = True
                return k
        if not valid:
            raise ValueError('Invalid time signature: {}/{}'.format(numerator, denominator))

    def get_note_list(self, with_drum=False) -> List[Tuple[int, int, int, int]]:
        '''
        Get all notes in the MultiTrack.

        Returns:
            List of tuples (onset, pitch, duration, velocity)
        '''
        all_notes = []
        for bar in self.bars:
            for inst_id, track in bar.tracks.items():
                if not with_drum and track.is_drum:
                    continue
                all_notes.extend(track.get_note_list())
        return all_notes



# 
'''
Author: Longshen Ou

Fixed: 
- calculation based on weighted histogram, where 1, 3, 5 are 2x important.
- 0 weight for off-scale notes, instead of -1
'''

import numpy as np
from typing import List, Tuple


key_profile = np.array([
    [ 1, -1,  1, -1,  1,  1, -1,  1, -1,  1, -1,  1], # C major
    [ 1,  1, -1,  1, -1,  1,  1, -1,  1, -1,  1, -1],
    [-1,  1,  1, -1,  1, -1,  1,  1, -1,  1, -1,  1],
    [ 1, -1,  1,  1, -1,  1, -1,  1,  1, -1,  1, -1],
    [-1,  1, -1,  1,  1, -1,  1, -1,  1,  1, -1,  1],
    [ 1, -1,  1, -1,  1,  1, -1,  1, -1,  1,  1, -1],
    [-1,  1, -1,  1, -1,  1,  1, -1,  1, -1,  1,  1],
    [ 1, -1,  1, -1,  1, -1,  1,  1, -1,  1, -1,  1],
    [ 1,  1, -1,  1, -1,  1, -1,  1,  1, -1,  1, -1],
    [-1,  1,  1, -1,  1, -1,  1, -1,  1,  1, -1,  1],
    [ 1, -1,  1,  1, -1,  1, -1,  1, -1,  1,  1, -1],
    [-1,  1, -1,  1,  1, -1,  1, -1,  1, -1,  1,  1],

    [ 1, -1,  1,  1, -1,  1, -1,  1,  1, -1,  1, -1], # A minor
    [-1,  1, -1,  1,  1, -1,  1, -1,  1,  1, -1,  1],
    [ 1, -1,  1, -1,  1,  1, -1,  1, -1,  1,  1, -1],
    [-1,  1, -1,  1, -1,  1,  1, -1,  1, -1,  1,  1],
    [ 1, -1,  1, -1,  1, -1,  1,  1, -1,  1, -1,  1],
    [ 1,  1, -1,  1, -1,  1, -1,  1,  1, -1,  1, -1],
    [-1,  1,  1, -1,  1, -1,  1, -1,  1,  1, -1,  1],
    [ 1, -1,  1,  1, -1,  1, -1,  1, -1,  1,  1, -1],
    [-1,  1, -1,  1,  1, -1,  1, -1,  1, -1,  1,  1],
    [ 1, -1,  1, -1,  1,  1, -1,  1, -1,  1, -1,  1],
    [ 1,  1, -1,  1, -1,  1,  1, -1,  1, -1,  1, -1],
    [-1,  1,  1, -1,  1, -1,  1,  1, -1,  1, -1,  1],
])


''' Apply weighting to key_profile '''
major_weight = [2, 0, 1.0, 0, 1.5, 1.0, 0, 1.5, 0, 1, 0, 1]
minor_weight = [2, 0, 1, 1.5, 0, 1.0, 0, 1.5, 1.0, 0, 1.0, 0]

for i in range(24):
    if i < 12:
        # circular right shift major weight
        shifted_major_weight = major_weight[-i:] + major_weight[:-i]

        # Position-wise multiplication
        for k in range(12):
            key_profile[i][k] *= shifted_major_weight[k]
    else:
        j = i - 12
        # circular shift minor weight
        shifted_minor_weight = minor_weight[-j:] + minor_weight[:-j]

        # Position-wise multiplication
        for k in range(12):
            key_profile[i][k] *= shifted_minor_weight[k]


def detect_key(note_list:List[Tuple[int, int, int, int]]):
    '''
    Determine the major/minor key and pitch shift needed for key normalization.

    Use this version instead. Others are deprecated.

    Args:
    - note_list: List of notes, each note is a tuple of (onset, pitch, duration, velocity)

    Returns:
    - is_major: True if major, False if minor
    - pitch_shift: The pitch shift needed for key normalization
    '''
    
    pitch_class_histogram = get_pitch_class_histogram(
        note_list, 
        normalize=True,
        use_duration=True, 
        use_velocity=True,
        note_has_onset=True,
    )

    pitches = [note[1] for note in note_list]
    min_pitch = min(pitches)
    max_pitch = max(pitches)

    pitch_shift, is_major = get_pitch_shift_from_pitch_histogram(
        pitch_class_histogram,
        min_pitch,
        max_pitch,
    )
    
    return is_major, pitch_shift


def get_pitch_class_histogram(
        notes:List[Tuple[int, int, int]], 
        normalize=True, 
        use_duration=True, 
        use_velocity=True,
        note_has_onset=False,
    ):
    '''
    Calculate the pitch class histogram of the notes.

    Args:
    - notes: List of notes, each note is a tuple of (pitch, duration, velocity)
    '''
    if note_has_onset:
        notes = [(note[1], note[2], note[3]) for note in notes]

    weights = np.ones(len(notes))
    # Assumes that duration and velocity have equal weight
    # (pitch, duration, velocity)
    if use_duration:
        weights *= [note[1] for note in notes]  # duration
    if use_velocity:
        weights *= [note[2] for note in notes]  # velocity
    histogram, _ = np.histogram([note[0] % 12 for note in notes], bins=np.arange(13), weights=weights,
                                density=normalize)
    if normalize:
        histogram_sum = histogram.sum()
        histogram /= (histogram_sum + (histogram_sum == 0))
    return histogram


def get_pitch_shift_from_pitch_histogram(histogram, min_pitch, max_pitch):
    key_candidate = np.dot(key_profile, histogram) # [24,]
    major_key_candidate = key_candidate[:12]
    minor_key_candidate = key_candidate[12:]

    major_index = np.argmax(major_key_candidate)
    minor_index = np.argmax(minor_key_candidate)

    major_score = major_key_candidate[major_index]
    minor_score = minor_key_candidate[minor_index]
    if major_score < minor_score:
        key_number = minor_index  # 小调
        is_major = False
        real_key = key_number
        pitch_shift = -3 - real_key  # 小调
    else:
        key_number = major_index  # 大调
        is_major = True
        real_key = key_number
        pitch_shift = 0 - real_key  

    # Make the shift to a nearer octave
    if pitch_shift > 6:
        pitch_shift -= 12
    elif pitch_shift < -6:
        pitch_shift += 12

    # Ensure the pitch shift is within the valid range
    while pitch_shift + min_pitch < 0:
        pitch_shift += 12
    while pitch_shift + max_pitch > 127:
        pitch_shift -= 12

    return pitch_shift, is_major